import {Component, ElementRef, ViewChild, OnInit, Input, OnDestroy, HostListener} from '@angular/core';

import {
  Label,
  MySprite, 
  tweenChange,
  ShapeRect,
  ShapeCircle,
  ShapeRoundRect

} from './Unit';
import {res, resAudio} from './resources';

import {Subject} from 'rxjs';
import {debounceTime} from 'rxjs/operators';

import TWEEN from '@tweenjs/tween.js';




@Component({
  selector: 'app-play',
  templateUrl: './play.component.html',
  styleUrls: ['./play.component.scss']
})
export class PlayComponent implements OnInit, OnDestroy {

  @ViewChild('canvas') canvas: ElementRef;
  @ViewChild('wrap') wrap: ElementRef;

  // 数据
  data;

  ctx;

  canvasWidth = 1280; // canvas实际宽度
  canvasHeight = 720; // canvas实际高度

  canvasBaseW = 1280; // canvas 资源预设宽度
  canvasBaseH = 720;  // canvas 资源预设高度

  mx; // 点击x坐标
  my; // 点击y坐标


  // 资源
  rawImages = new Map(res);
  rawAudios = new Map(resAudio);

  images = new Map();

  animationId: any;
  winResizeEventStream = new Subject();

  audioObj = {};

  renderArr;
  mapScale = 1;

  canvasLeft;
  canvasTop;

  //间隔宽度
  interval_width;

  //间隔高度
  interval_height;

  //单词数组
  wordSprArr = [];

  //标题背景
  title_bg

  saveKey = 'ym_018';

  @HostListener('window:resize', ['$event'])
  onResize(event) {
    this.winResizeEventStream.next();
  }


  ngOnInit() {

    this.data = {};

    // 获取数据
    const getData = (<any> window).courseware.getData;
    getData((data) => {

      if (data && typeof data == 'object') {
        this.data = data;
      }

      console.log('data:' , data);

      // 初始化 各事件监听
      this.initListener();

      // 若无数据 则为预览模式 需要填充一些默认数据用来显示
      this.initDefaultData();

      // 初始化 音频资源
      this.initAudio();
      // 初始化 图片资源
      this.initImg();
      // 开始预加载资源
      this.load();

    }, this.saveKey);

  }

  getDefaultTitle(){
    return {
      t_val:'',
      t_font_color:'#000000',
      t_font:'BRLNSDB',
      t_font_size:40
    };
  }

  ngOnDestroy() {
    window['curCtx'] = null;
    window.cancelAnimationFrame(this.animationId);
  }


  load() {

    // 预加载资源
    this.loadResources().then(() => {
      window["air"].hideAirClassLoading(this.saveKey, this.data);
      this.init();
      this.update();
    });
  }


  init() {

    this.initCtx();
    this.initData();
    this.initView();
  }

  initCtx() {
    this.canvasWidth = this.wrap.nativeElement.clientWidth;
    this.canvasHeight = this.wrap.nativeElement.clientHeight;
    this.canvas.nativeElement.width = this.wrap.nativeElement.clientWidth;
    this.canvas.nativeElement.height = this.wrap.nativeElement.clientHeight;


    this.ctx = this.canvas.nativeElement.getContext('2d');
    this.canvas.nativeElement.width = this.canvasWidth;
    this.canvas.nativeElement.height = this.canvasHeight;

    window['curCtx'] = this.ctx;
  }

  updateItem(item) {
    if (item) {
      item.update();
    }
  }

  updateArr(arr) {
    if (!arr) {
      return;
    }
    for (let i = 0; i < arr.length; i++) {
      arr[i].update(this);
    }
  }

  initListener() {

    this.winResizeEventStream
      .pipe(debounceTime(500))
      .subscribe(data => {
        this.renderAfterResize();
      });


    // ---------------------------------------------
    const setParentOffset = () => {
      const rect = this.canvas.nativeElement.getBoundingClientRect();
      this.canvasLeft = rect.left;
      this.canvasTop = rect.top;
    };
    const setMxMyByTouch = (event) => {
      if (event.touches.length <= 0) {
        return;
      }
      if (this.canvasLeft == null) {
        setParentOffset();
      }
      this.mx = event.touches[0].pageX - this.canvasLeft;
      this.my = event.touches[0].pageY - this.canvasTop;
    };

    const setMxMyByMouse = (event) => {
      this.mx = event.offsetX;
      this.my = event.offsetY;
    };
    // ---------------------------------------------


    let firstTouch = true;

    const touchDownFunc = (e) => {
      if (firstTouch) {
        firstTouch = false;
        removeMouseListener();
      }
      setMxMyByTouch(e);
      this.mapDown(e);
    };
    const touchMoveFunc = (e) => {
      setMxMyByTouch(e);
      this.mapMove(e);
    };
    const touchUpFunc = (e) => {
      setMxMyByTouch(e);
      this.mapUp(e);
    };

    const mouseDownFunc = (e) => {
      if (firstTouch) {
        firstTouch = false;
        removeTouchListener();
      }
      setMxMyByMouse(e);
      this.mapDown(e);
    };
    const mouseMoveFunc = (e) => {
      setMxMyByMouse(e);
      this.mapMove(e);
    };
    const mouseUpFunc = (e) => {
      setMxMyByMouse(e);
      this.mapUp(e);
    };


    const element = this.canvas.nativeElement;

    const addTouchListener = () => {
      element.addEventListener('touchstart', touchDownFunc);
      element.addEventListener('touchmove', touchMoveFunc);
      element.addEventListener('touchend', touchUpFunc);
      element.addEventListener('touchcancel', touchUpFunc);
    };
    const removeTouchListener = () => {
      element.removeEventListener('touchstart', touchDownFunc);
      element.removeEventListener('touchmove', touchMoveFunc);
      element.removeEventListener('touchend', touchUpFunc);
      element.removeEventListener('touchcancel', touchUpFunc);
    };

    const addMouseListener = () => {
      element.addEventListener('mousedown', mouseDownFunc);
      element.addEventListener('mousemove', mouseMoveFunc);
      element.addEventListener('mouseup', mouseUpFunc);
    };
    const removeMouseListener = () => {
      element.removeEventListener('mousedown', mouseDownFunc);
      element.removeEventListener('mousemove', mouseMoveFunc);
      element.removeEventListener('mouseup', mouseUpFunc);
    };

    addMouseListener();
    addTouchListener();
  }


  playAudio(key, now = false, callback = null) {

    const audio = this.audioObj[key];
    if (audio) {
      if (now) {
        audio.pause();
        audio.currentTime = 0;
      }

      if (callback) {
        audio.onended = () => {
          callback();
        };
      }
      audio.play();
    }
  }


  pauseAudio(key, callback = null){

    const audio = this.audioObj[key];
    if (audio) {

      if (callback) {
        audio.onended = () => {
          callback();
        };
      }
      audio.pause();
    }
  }


  loadResources() {
    const pr = [];
    this.rawImages.forEach((value, key) => {// 预加载图片

      const p = this.preload(value)
        .then(img => {
          this.images.set(key, img);
        })
        .catch(err => console.log(err));

      pr.push(p);
    });

    this.rawAudios.forEach((value, key) => {// 预加载音频

      const a = this.preloadAudio(value)
        .then(() => {
          // this.images.set(key, img);
        })
        .catch(err => console.log(err));

      pr.push(a);
    });
    return Promise.all(pr);
  }

  preload(url) {
    return new Promise((resolve, reject) => {
      const img = new Image();
      // img.crossOrigin = "anonymous";
      img.onload = () => resolve(img);
      img.onerror = reject;
      img.src = url;
    });
  }

  preloadAudio(url) {
    return new Promise((resolve, reject) => {
      const audio = new Audio();
      audio.oncanplay = (a) => {
        resolve();
      };
      audio.onerror = () => {
        reject();
      };
      audio.src = url;
      audio.load();
    });
  }


  renderAfterResize() {
    this.canvasWidth = this.wrap.nativeElement.clientWidth;
    this.canvasHeight = this.wrap.nativeElement.clientHeight;

    // 若无数据 则为预览模式 需要填充一些默认数据用来显示
    this.initDefaultData();

    this.init();
  }

  /**
   * 检查鼠标点击目标视图里
   * @param target 目标视图
   */
  checkClickTarget(target) {

    const rect = target.getBoundingBox();

    if (this.checkPointInRect(this.mx, this.my, rect)) {
      return true;
    }
    return false;
  }


  checkPointInRect(x, y, rect) {
    if (x >= rect.x && x <= rect.x + rect.width) {
      if (y >= rect.y && y <= rect.y + rect.height) {
        return true;
      }
    }
    return false;
  }

  addUrlToAudioObj(key, url = null, vlomue = 1, loop = false, callback = null, callbackParam = null) {

    console.log('add audio', key);
    if(!key){
      return;
    }

    const audioObj = this.audioObj;

    if (url == null) {
      url = key;
    }

    this.rawAudios.set(key, url);

    const audio = new Audio();
    audio.src = url;
    audio.load();
    audio.loop = loop;
    audio.volume = vlomue;

    audioObj[key] = audio;

    audio.addEventListener('ended', function () {  
      console.log('play ended');
      if (callback) {
        callback(callbackParam);
      }
    }, false);
  }

  addUrlToImages(url) {
    this.rawImages.set(url, url);
  }






  // ======================================================编写区域==========================================================================





  /**
   * 添加默认数据 便于无数据时的展示
   */
  initDefaultData() {

    if(!this.data.title){
      this.data.title = this.getDefaultTitle();
    }

    if(!this.data.words){
      this.data.words = [];
    }

    this.wordSprArr = [];

    //间隔宽度
    this.interval_width = 250;

    //间隔高度
    this.interval_height = 60;

    this.title_bg = {};
  }


  /**
   * 添加预加载图片
   */
  initImg() {

  }

  /**
   * 添加预加载音频
   */
  initAudio() {

    // 音效
    this.addUrlToAudioObj('click', this.rawAudios.get('click'), 0.3);

    //标题发音
    this.addUrlToAudioObj(this.data.title.t_audio_url);

    //单词发音
    for(let i = 0; i < this.data.words.length; ++ i){
      let word = this.data.words[i];
      this.addUrlToAudioObj(word.word_audio_url);
    }
  }

  /**
   * 初始化数据
   */
  initData() {

    const sx = this.canvasWidth / this.canvasBaseW;
    const sy = this.canvasHeight / this.canvasBaseH;
    const s = Math.min(sx, sy);
    this.mapScale = s;

    this.interval_height = this.interval_height * s;

    this.interval_width = this.interval_width * s;

    this.renderArr = [];

  }



  /**
   * 初始化视图
   */
  initView() {

    //初始化标题
    this.initTitle();

    //初始化单词
    this.initWord();
  }


  /**
   * 初始化标题
   */
  initTitle(){
    const title_bg = new ShapeRoundRect();
    title_bg.y = -24;
    title_bg.fillColor = "#975833";
    this.title_bg = title_bg;
    this.renderArr.push(title_bg);

    let title_content = this.data.title;
    console.log('title', title_content);

    //标题1
    let title1;

    //标题2
    let title2;

    if(title_content){

      //标题1
      title1 = new Label();    
      title1.text = "Lesson";
      title1.textAlign = 'center';
      title1.fontSize = 40;
      title1.fontName = "BRLNSDB";
      title1.fontColor = "	#FFFFFF";
      title1.refreshSize();
      title_bg.addChild(title1);

      title1.y = (title1.height + 10)/2;

      //标题2
      title2 = new Label();    
      title2.text = title_content.t_val;
      title2.textAlign = 'center';
      title2.fontSize = 40;
      title2.fontName = "BRLNSDB";
      title2.fontColor = "#975833";
      title2.refreshSize();
    }

    //标题2背景
    let title2_bg = new ShapeCircle();
    title2_bg.setRadius(30);
    title2_bg.fillColor="#F8C224";
    title2_bg.addChild(title2);

    //将标题1添加到标题背景里
    title_bg.addChild(title2_bg);

    //标题宽度
    let titleTotalWidth = title1.width + 10 + title2_bg.width;

    let title_bg_width = titleTotalWidth + 50;

    this.title_bg.setSize(title_bg_width, 110, 24);

    this.title_bg.x = (this.canvasWidth -  title_bg_width)/ 2;

    //重新设置字母（组合）的坐标
    let titleStartX = (title_bg_width - titleTotalWidth)/2;

    //标题 1坐标
    title1.y = this.title_bg.height /2 + 12
    title1.x = titleStartX + title1.width/2;
    //父类拉伸后，子类不拉伸
    title1.scaleX /= this.title_bg.scaleX;

    titleStartX += title1.width + 20;

    //标题 2坐标
    title2_bg.y = this.title_bg.height /2 + 12
    title2_bg.x = titleStartX + title2_bg.width/2;
    //父类拉伸后，子类不拉伸
    title2_bg.scaleX /= this.title_bg.scaleX;

    this.title_bg = title_bg;

    this.renderArr.push(title_bg);
  }

  /**
   * 初始化单词
   */
  initWord(){

    let maxWidth = 0;

    let maxHeiht = 0;

    let v_word_arr = [];

    //循环生成单词视图
    for(let i = 0; i < this.data.words.length; ++ i){

      let word = this.data.words[i];

      let v_word = new ShapeRect();
      v_word.alpha=0;
      v_word.setScaleXY(this.mapScale);
    
      this.renderArr.push(v_word);
      this.wordSprArr.push(v_word);
       
      //单词
      let word_val = this.getWordValSprite(word, v_word);

      if(maxWidth < v_word.width){
        maxWidth = v_word.width;
      }

      maxHeiht = v_word.height;
    }

    console.log('word_spr_arr', this.wordSprArr);

    //行数
    let rows = 0;

    //列数
    let cols = 0;

    //单词个数
    let wordNum = this.wordSprArr.length;

    if(wordNum < 4){
      rows = 1;

      cols = wordNum;
    }
    else if(wordNum == 4){
      rows = 2;

      cols = 2;
    }
    else {

      cols = 3;

      rows = Math.floor(wordNum/3);

      if(wordNum % 3 != 0){
        rows += 1;
      }
    }

    //总宽度
    let totalWidth = 0;

    //总高度
    let totalHeight = rows * maxHeiht + (rows - 1) * this.interval_height;

    //起始Y坐标
    let startY = (this.canvasHeight - totalHeight) / 2;

    console.log('canvasHeight', this.canvasHeight);

    console.log('title_bg', this.title_bg.height);
    console.log('totalHeight', totalHeight);

    let row_max_width = cols * maxWidth + (cols - 1) * this.interval_width;

    //起始X坐标
    let startX = (this.canvasWidth - row_max_width)/2;

    //每行起始坐标
    let row_startX = startX;
    
    for(let i = 0; i < wordNum; ++ i){

      let index = i + 1;
      let word_spreat = this.wordSprArr[i];

      word_spreat.y = startY;

      word_spreat.x = row_startX;

      // 如果不是行末尾，增加空格
      if(index % cols != 0 && index != wordNum){

        row_startX += maxWidth + this.interval_width;
      }
      // 到最后一个单词
      else if(index % cols != 0 && index == wordNum){

        
      }
      // 每行显示n个单词
      else if(index % cols == 0){
        //每一行最后一个单词，高度起始坐标增加
        startY += word_spreat.height + this.interval_height;

        //下一行X坐标复原
        row_startX = startX;
      }

      console.log('word_spr', word_spreat);
    }
  }

  /**
   * 获取单词值视图
   * @param word 
   * @param word_bg 
   */
  getWordValSprite(word, word_bg){

    let word_val = new Label();
    word_val.text = word.word_val;
    word_val.textAlign = 'center';
    word_val.fontSize = 110;
    word_val.fontName = "GOTHICB";
    word_val.fontColor = "#674023";
    word_val.refreshSize();
    word_bg.addChild(word_val);

    word_val.x = word_val.width/2;

    word_val.y = word_val.height/2;

    word_bg.setSize(word_val.width, word_val.height);

    return word_val;
  }

  mapDown(event) {


    for(let i = 0; i < this.wordSprArr.length; ++ i){

      let wordSpr = this.wordSprArr[i];
      
      if(this.checkClickTarget(wordSpr)){
        
        this.playAudio(this.data.words[i].word_audio_url);
        return;
      }
    }

    if(this.checkClickTarget(this.title_bg)){
      console.log('click title');
      this.playAudio(this.data.title.t_audio_url);
      return;
    }
  }

  mapMove(event) {

  }

  mapUp(event) {

  }



  update() {

    // ----------------------------------------------------------
    this.animationId = window.requestAnimationFrame(this.update.bind(this));
    // 清除画布内容
    this.ctx.clearRect(0, 0, this.canvasWidth, this.canvasHeight);
    // tween 更新动画
    TWEEN.update();
    // ----------------------------------------------------------



    this.updateArr(this.renderArr);


  }

}
